#version 450 core

const int MAX_LIGHTS = 64;
const float Epsilon = 1e-10;
const int LIGHT_SAMPLE_RATE = 16;
const int TEXTURE_SAMPLE_RATE = 12;

struct PointLight {
    vec3 position;
    vec3 intensity;
    vec3 attenuation;
};

struct DirectionalLight {
    vec3 direction;
    vec3 intensity;
};

struct SpotLight {
    vec3 position;
    vec3 direction;
    vec3 intensity;
    vec3 attenuation;
    float angle;
};

in vec3 fragPos;
in vec3 fragNormal;
in vec2 fragTexCoordinate;

layout(location=0) out vec4 color;
layout(location=1) out vec4 normalColor;

// Texture
uniform sampler2D diffuseTexture;
uniform bool useDiffuseTexture;
uniform sampler2D specularTexture;
uniform bool useSpecularTexture;
// Color of the object (only used if texture is not set)
uniform vec4 inColor;
uniform vec4 inSpecularColor;
// ambient lighting factor
uniform float ka;
// diffuse lighting factor
uniform float kd;
// specular lighting factor
uniform float ks;
// specular factor
uniform float alpha;
// camera position
uniform vec3 eyePos;
uniform PointLight pointLight[MAX_LIGHTS];
uniform int numberOfPointLights;
uniform DirectionalLight directionalLight[MAX_LIGHTS];
uniform int numberOfDirectionalLights;
uniform SpotLight spotLight[MAX_LIGHTS];
uniform int numberOfSpotLights;
uniform bool useEnvironmentMapping;
// environment map factor
uniform float ke;
uniform samplerCube skybox;
// brightness factor
uniform float brightness;

vec3 calculateDiffuseFromDirectionalLight(DirectionalLight directionalLight);
vec3 calculateSpecularFromDirectionalLight(DirectionalLight directionalLight);
vec3 calculateDiffuseFromPointLight(PointLight pointLight);
vec3 calculateSpecularFromPointLight(PointLight pointLight);
vec3 calculateDiffuseFromSpotLight(SpotLight spotLight);
vec3 calculateSpecularFromSpotLight(SpotLight spotLight);

vec3 rgbToHsv(vec3 RGB);
vec3 hsvtoRgb(vec3 HSV);

vec3 normal;

void main() {
    normal = normalize(fragNormal);
    normalColor = abs(vec4(normal, 1.0f));
    // Ambient light
    vec3 lightIntensity = vec3(1.0f) * ka;
    vec3 specularIntensity = vec3(0.0f);

    // Point lights
    for (int i = 0; i < numberOfPointLights; i++) {
        lightIntensity += calculateDiffuseFromPointLight(pointLight[i]);
        specularIntensity += calculateSpecularFromPointLight(pointLight[i]);
    }

    // Directional lights
    for (int i = 0; i < numberOfDirectionalLights; i++) {
        lightIntensity += calculateDiffuseFromDirectionalLight(directionalLight[i]);
        specularIntensity += calculateSpecularFromDirectionalLight(directionalLight[i]);
    }

    // Spot lights
    for (int i = 0; i < numberOfSpotLights; i++) {
        lightIntensity += calculateDiffuseFromSpotLight(spotLight[i]);
        specularIntensity += calculateSpecularFromSpotLight(spotLight[i]);
    }

    vec4 diffuseColor = inColor;
    vec4 specularColor = inSpecularColor;

    lightIntensity = floor(lightIntensity * LIGHT_SAMPLE_RATE) / LIGHT_SAMPLE_RATE;

	if (useDiffuseTexture) {
        // convert RGB color to HSV color space to discretize the brightness
        diffuseColor = texture(diffuseTexture, fragTexCoordinate);
        vec3 rgb_values = vec3(diffuseColor.r, diffuseColor.g, diffuseColor.b);
        vec3 hsv = rgbToHsv(rgb_values);
        hsv.z = floor(hsv.z * TEXTURE_SAMPLE_RATE) / TEXTURE_SAMPLE_RATE;
        diffuseColor = vec4(hsvtoRgb(hsv), diffuseColor.w);
    }

	if (useSpecularTexture) {
        specularColor = texture(specularTexture, fragTexCoordinate);
    }
    vec4 environmentMapColor = vec4(0.0f);
    if (useEnvironmentMapping) {
        // Directional vector from viewer
        vec3 V = normalize(vec3(fragPos) - eyePos);
        // Reflection
        vec3 R = reflect(V, normal);
        environmentMapColor += texture(skybox, R);

        vec3 rgb_values = vec3(environmentMapColor.rgb);
        vec3 hsv = rgbToHsv(rgb_values);
        hsv.z = floor(hsv.z * TEXTURE_SAMPLE_RATE) / TEXTURE_SAMPLE_RATE;
        environmentMapColor = vec4(hsvtoRgb(hsv), environmentMapColor.w);
    }

    color = diffuseColor * vec4(lightIntensity, 1.0f) * brightness + specularColor * vec4(specularIntensity, 1.0f) * brightness + ke * environmentMapColor;
}

vec3 calculateDiffuseFromDirectionalLight(DirectionalLight directionalLight) {
    // Directional vector to light source
    vec3 L = normalize(-directionalLight.direction);
    return directionalLight.intensity * kd * max(0.0f, dot(L, normal));
}

vec3 calculateSpecularFromDirectionalLight(DirectionalLight directionalLight) {
    // Directional vector to light source
    vec3 L = normalize(-directionalLight.direction);
    // Reflection
    vec3 R = reflect(-L, normal);
    // Directional vector to viewer
    vec3 V = normalize(eyePos - vec3(fragPos));
    return directionalLight.intensity * ks * pow(max(0.0f, dot(R, V)), alpha);
}

vec3 calculateDiffuseFromPointLight(PointLight pointLight) {
    // Directional vector to light source
    vec3 L = normalize(pointLight.position - vec3(fragPos));
    float attenuationFactor = pointLight.attenuation.x + pointLight.attenuation.y * distance(pointLight.position, vec3(fragPos)) + pointLight.attenuation.z * pow(distance(pointLight.position, vec3(fragPos)), 2.0f);
    return pointLight.intensity / attenuationFactor * kd * max(0.0f, dot(L, normal));
}

vec3 calculateSpecularFromPointLight(PointLight pointLight) {
    vec3 L = normalize(pointLight.position - vec3(fragPos));
    // Reflection
    vec3 R = reflect(-L, normal);
    // Directional vector to viewer
    vec3 V = normalize(eyePos - vec3(fragPos));
    float attenuationFactor = pointLight.attenuation.x + pointLight.attenuation.y * distance(pointLight.position, vec3(fragPos)) + pointLight.attenuation.z * pow(distance(pointLight.position, vec3(fragPos)), 2.0f);
    return pointLight.intensity / attenuationFactor * ks * pow(max(0.0f, dot(R, V)), alpha);
}

vec3 calculateDiffuseFromSpotLight(SpotLight spotLight) {
    // Directional vector to light source
    vec3 L = normalize(spotLight.position - vec3(fragPos));
    float attenuationFactor = spotLight.attenuation.x + spotLight.attenuation.y * distance(spotLight.position, vec3(fragPos)) + spotLight.attenuation.z * pow(distance(spotLight.position, vec3(fragPos)), 2.0f);

    vec3 spotDir = normalize(-spotLight.direction);
    float intensity = smoothstep(0.0f, 0.02f, max(dot(L, spotDir) - cos(spotLight.angle), 0.0f));

    return intensity * spotLight.intensity / attenuationFactor * kd * max(dot(normal, L), 0.0f);
}

vec3 calculateSpecularFromSpotLight(SpotLight spotLight) {
    vec3 L = normalize(spotLight.position - vec3(fragPos));
    // Reflection
    vec3 R = reflect(-L, normal);
    // Directional vector to viewer
    vec3 V = normalize(eyePos - vec3(fragPos));
    float attenuationFactor = spotLight.attenuation.x + spotLight.attenuation.y * distance(spotLight.position, vec3(fragPos)) + spotLight.attenuation.z * pow(distance(spotLight.position, vec3(fragPos)), 2.0f);

    vec3 spotDir = normalize(-spotLight.direction);
    float intensity = smoothstep(0.0f, 0.02f, max(dot(L, spotDir) - cos(spotLight.angle), 0.0f));

    return intensity * spotLight.intensity / attenuationFactor * ks * pow(max(0.0f, dot(R, V)), alpha);
}

// https://stackoverflow.com/questions/53879537/increase-the-intensity-of-texture-in-shader-code-opengl
vec3 rgbToHsv(vec3 RGB) {
    vec4  P   = (RGB.g < RGB.b) ? vec4(RGB.bg, -1.0, 2.0/3.0) : vec4(RGB.gb, 0.0, -1.0/3.0);
    vec4  Q   = (RGB.r < P.x) ? vec4(P.xyw, RGB.r) : vec4(RGB.r, P.yzx);
    float C   = Q.x - min(Q.w, Q.y);
    float H   = abs((Q.w - Q.y) / (6.0 * C + Epsilon) + Q.z);
    vec3  HCV = vec3(H, C, Q.x);
    float S   = HCV.y / (HCV.z + Epsilon);
    return vec3(HCV.x, S, HCV.z);
}

vec3 hsvtoRgb(vec3 HSV) {
    float H   = HSV.x;
    float R   = abs(H * 6.0 - 3.0) - 1.0;
    float G   = 2.0 - abs(H * 6.0 - 2.0);
    float B   = 2.0 - abs(H * 6.0 - 4.0);
    vec3  RGB = clamp( vec3(R,G,B), 0.0, 1.0 );
    return ((RGB - 1.0) * HSV.y + 1.0) * HSV.z;
}